/********
 * gtksu - a lightweight GTK gui for su (for root only)
 * (C) James Budiono 2012
 * Note: does not claim to be secure or correct!
 ********/
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <sys/types.h>
#include <pwd.h>
#include <grp.h>
#include <shadow.h>
#include <crypt.h>

#include <gtk/gtk.h>
#include <locale.h>

#include "gtksu.h"

char *run_as_user="nobody";
extern char **environ;
char *root_password;
int ipc[2];
char *good_code="good";
char *bad_code="bad";

int get_root_password() {
	struct passwd *pw = getpwuid(0); //root
	if (!pw) return FALSE;
		
	struct spwd *spwd = getspnam(pw->pw_name);
	if (!spwd) return FALSE;
	
	root_password = strdup(spwd->sp_pwdp);
	return TRUE;
}

int init(int argc, char *argv[]) {
	pid_t pid;
	char buf[32];
	
	if ( ! get_root_password() ) return FALSE;
	
	/* create pipe */
	if ( pipe(ipc) ) return FALSE;
	
	pid=fork();
	if (pid == (pid_t)-1) return FALSE;
	
	if (pid) {
		// parent - change identity to root, read pipe and wait for instruction
		initgroups("root", 0);
		endgrent(); 
		setgid(0);
		setuid(0);
		
		read (ipc[0], buf, sizeof(buf));
		if ( strncmp (buf, good_code, sizeof(good_code)) == 0 ) execvp(argv[1],&argv[1]);
		else exit(1);
	} else {
		// child - drop privilege (so that GTK does not complain) and continue to validate the password
		struct passwd *pw = getpwnam(run_as_user);
		if (!pw) return FALSE;
		setgid (pw->pw_gid);		
		setuid (pw->pw_uid);
		return TRUE;		
	}
}

int validate_password (char *password) {
	char *encrypted;
	encrypted = crypt(password, root_password);
	return (strcmp(encrypted, root_password) == 0);
}

int confirm_password(int argc, char *argv[]) {
    GtkBuilder *builder;
    GError     *error = NULL;
    
	GtkDialog  *error_dialog;    
    GtkDialog  *password_dialog;
    GtkEntry   *password_entry;
    
    char title[1024];
    int  result;
    char *password;

	setlocale (LC_ALL, "");
	bindtextdomain ("gtksu", "/usr/share/locale");
	textdomain ("gtksu");
	
    gtk_init( &argc, &argv );
    builder = gtk_builder_new();    
    
    //if( ! gtk_builder_add_from_file( builder, GTKSU_GUI, &error ) )
    if( ! gtk_builder_add_from_string( builder, GTKSU_GUI, -1, &error ) )
    {
        g_error ( "%s", error->message );
        return( FALSE );
    }
    
    password_dialog = (GtkDialog *) gtk_builder_get_object ( builder, "password_dialog" ); 
    password_entry	= (GtkEntry *)  gtk_builder_get_object ( builder, "password" ); 
    g_object_unref( G_OBJECT( builder ) );
    
    snprintf ( title, sizeof(title), "Run %s as Root", argv[1] );
	gtk_window_set_title ( (GtkWindow *) password_dialog, title );    
	
	while (TRUE) {
		result = gtk_dialog_run ( password_dialog );
		//g_debug ( "dialog result %d\n", result );
		if (result == 1) {
			password = (char *) gtk_entry_get_text ( password_entry );
			//g_debug ( "password %s\n", password );
			if ( validate_password (password) ) return TRUE;
			else {
				error_dialog = (GtkDialog *) gtk_message_dialog_new (NULL,
												  GTK_DIALOG_DESTROY_WITH_PARENT,
												  GTK_MESSAGE_ERROR,
												  GTK_BUTTONS_CLOSE,
												  "Wrong password.");
				 gtk_window_set_title ( (GtkWindow *) error_dialog, "Error !" );
				 gtk_dialog_run ( error_dialog );
				 gtk_widget_destroy ((GtkWidget *) error_dialog);
				 gtk_widget_grab_focus ( (GtkWidget *) password_entry );
			}
		} else break;
	}	
    gtk_widget_destroy ( (GtkWidget *) password_dialog );
	return FALSE;
}

void main(int argc, char *argv[]) {
	if (argc < 2) {
		printf ( "Usage: %s command [args]\n", argv[0] );
		exit (1);
	}
	
	if ( ! init(argc, argv) ) exit(1);
	
	if ( confirm_password(argc, argv) ) {
		write (ipc[1], good_code, sizeof(good_code)); 
		exit(0);
	} else  {
		write (ipc[1], bad_code, sizeof(bad_code)); 
		exit(1);
	}
}